mod river_protocols;

use river_protocols::zriver_seat_status_v1::ZriverSeatStatusV1;
use river_protocols::{
    zriver_output_status_v1, zriver_seat_status_v1,
    zriver_status_manager_v1::ZriverStatusManagerV1,
};
use serde::ser::{SerializeSeq, SerializeStruct, Serializer};
use serde::Serialize;
use wayland_client::protocol::wl_registry::WlRegistry;
use std::collections::{BTreeMap, HashMap};
use wayland_client::protocol::{wl_output, wl_output::WlOutput, wl_seat,
    wl_seat::WlSeat};
use wayland_client::{Attached, DispatchData, Display, GlobalEvent, GlobalManager, Main};

#[derive(Debug, Default)]
struct Flags {
    once: bool,
    seat: Option<String>,
}


#[derive(Clone, Default)]
struct Tags(u32);

impl Serialize for Tags {
    fn serialize<S>(&self, serializer: S) -> Result<S::Ok, S::Error>
    where
        S: Serializer,
    {
        let mut seq = serializer.serialize_seq(Some(self.0.count_ones() as usize))?;
        for i in 0..32 {
            if self.0 >> i & 1 == 1 {
                seq.serialize_element(&(i + 1))?;
            }
        }
        seq.end()
    }
}


#[derive(Clone, serde::Serialize, Default)]
struct OutputGeometry {
    x: i32,
    y: i32,
    width: i32,
    height: i32,
    scale: i32,
    refresh: f32,
}

#[derive(Clone, Default, serde::Serialize)]
struct Output {
    #[serde(skip_serializing)]
    name: String,
    title: String,
    geometry: OutputGeometry,
    focused_tags: Tags,
    urgent_tags: Tags,
    populated_tags: Vec<u32>,
    // for serde serialization
    focused: bool,
    gdk_id: u32,
}

#[derive(Default)]
struct ReadStatus {
    wl_output: bool,
    wl_seat: bool,
    river_seat: bool,
    river_output: bool,
}

struct Env {
    status_manager: Option<Main<ZriverStatusManagerV1>>,
    flags: Flags,
    read_status: ReadStatus,
    // I don't know very much about wayland, and it shows
    output_id_map: HashMap<u32, u32>,
    layout: Option<String>,
    mode: Option<String>,
    focused_output: Option<u32>,
    outputs: BTreeMap<u32, Output>,
}

impl Env {
    fn new() -> Env {
        Env {
            status_manager: None,
            flags: configuration(),
            read_status: Default::default(),
            output_id_map: Default::default(),
            layout: None,
            mode: None,
            focused_output: None,
            outputs: Default::default(),
        }
    }

    fn fmt(&self) {
        println!("{}", serde_json::to_string(self).unwrap());
    }

    fn get_output(&mut self, id: u32) -> &mut Output {
        if !self.outputs.contains_key(&id) {
            self.outputs.insert(id, Default::default());
        }
        self.outputs.get_mut(&id).unwrap()
    }

    fn is_empty(&self) -> bool {
        self.layout.is_none() && self.mode.is_none() && self.outputs.len() == 0
    }

    fn is_first_read_done(&self) -> bool{
        self.read_status.wl_output && self.read_status.wl_seat &&
            self.read_status.river_seat && self.read_status.river_output
    }
}

impl Serialize for Env {
    fn serialize<S>(&self, serializer: S) -> Result<S::Ok, S::Error>
    where S: Serializer {
        let mut state = serializer.serialize_struct("Env", 9)?;
        if let Some(layout) = self.layout.as_ref() {
            state.serialize_field("layout", layout)?;
        } else {
            state.skip_field("layout")?;
        }
        if let Some(mode) = self.mode.as_ref() {
            state.serialize_field("mode", mode)?;
        } else {
            state.skip_field("mode")?;
        }
        let mut outputs: HashMap<String, Output> = HashMap::new();
        let mut gdk_id = 0;
        for (output_id, output) in &self.outputs {
            let mut new = output.clone();
            new.focused = if let Some(focused_id) = self.focused_output {
                *output_id == focused_id
            } else {
                false
            };
            new.gdk_id = gdk_id;
            outputs.insert(new.name.clone(), new);
            gdk_id += 1;
        }
        state.serialize_field("outputs", &outputs)?;
        state.end()
    }
}

fn handle_river_seat_status(
    _status: Main<ZriverSeatStatusV1>, event: zriver_seat_status_v1::Event,
    mut env: DispatchData
) {
    if let Some(env) = env.get::<Env>() {
        env.read_status.river_seat = true;
        match event {
            zriver_seat_status_v1::Event::FocusedOutput { output } => {
                env.focused_output = Some(output.as_ref().id());
            },
            zriver_seat_status_v1::Event::FocusedView { title } => {
                if let Some(focused_output) = env.focused_output {
                    env.get_output(focused_output).title = title;
                }
            },
            zriver_seat_status_v1::Event::Mode { name } => {
                env.mode = Some(name);
            },
            _ => {},
        }
    }
}

fn handle_seat_event(
    seat: Main<WlSeat>, event: wl_seat::Event, mut env: DispatchData
) {
    if let Some(env) = env.get::<Env>() {
        env.read_status.wl_seat = true;
        match event {
            wl_seat::Event::Name { name } => {
                if env.flags.seat.is_none()
                    || name == *env.flags.seat.as_ref().unwrap() {
                        if let Some(status_manager) = &env.status_manager {
                            let seat_status =
                                status_manager.get_river_seat_status(&seat);
                            seat_status.quick_assign(handle_river_seat_status);
                        }
                    }
            },
            _ => {},
        }
    }
}

fn handle_river_output_status(
    output: &Main<WlOutput>, event: zriver_output_status_v1::Event,
    mut env: DispatchData
) {
    if let Some(env) = env.get::<Env>() {
        env.read_status.river_output = true;
        let output_id = output.as_ref().id();
        match event {
            zriver_output_status_v1::Event::FocusedTags {
                tags: focused_tags,
            } => env.get_output(output_id).focused_tags = Tags(focused_tags),
            zriver_output_status_v1::Event::ViewTags {
                tags,
            } => env.get_output(output_id).populated_tags = tags[0..]
                .chunks(4)
                .map(|s| {
                    let buf = [s[0], s[1], s[2], s[3]];
                    let tagmask = u32::from_le_bytes(buf);
                    for i in 0..32 {
                        if 1 << i == tagmask {
                            return 1 + i;
                        }
                    }
                    0
                }).collect(),
            zriver_output_status_v1::Event::UrgentTags {
                tags,
            } => env.get_output(output_id).urgent_tags = Tags(tags),
            zriver_output_status_v1::Event::LayoutName {
                name,
            } => env.layout = Some(name),
            _ => {},
        }
    }
    
}

fn handle_output_event(
    output: Main<WlOutput>, event: wl_output::Event, mut env: DispatchData
) {
    let output_id = output.as_ref().id();
    if let Some(env) = env.get::<Env>() {
        env.read_status.wl_output = true;
        match event {
            wl_output::Event::Geometry {
                x,
                y,
                physical_width: _,
                physical_height: _,
                subpixel: _,
                make: _,
                model: _,
                transform: _,
            } => {
                if let Some(status_manager) = &env.status_manager {
                    let output_status =
                        status_manager.get_river_output_status(&output);
                    output_status.quick_assign(move |_, event, env| {
                        handle_river_output_status(&output, event, env);
                    });
                }
                let entry = env.get_output(output_id);
                entry.geometry.x = x;
                entry.geometry.y = y;
            },
            wl_output::Event::Mode { flags: _, width, height, refresh } => {
                let entry = env.get_output(output_id);
                entry.geometry.width = width;
                entry.geometry.height = height;
                entry.geometry.refresh = refresh as f32 / 1000.0f32;
            },
            wl_output::Event::Name {
                name,
            } => env.get_output(output_id).name = name,
            wl_output::Event::Scale {
                factor
            } => env.get_output(output_id).geometry.scale = factor,
            _ => {},
        }
    }
}

fn global_manager_callback(
    event: GlobalEvent, registry: Attached<WlRegistry>, mut env: DispatchData
) {
    match event {
        GlobalEvent::New { id, interface, version } => {
            match interface.as_str() {
                "wl_output" if version >= 3 => {
                    let output: Main<WlOutput> = registry.bind(version, id);
                    if let Some(env) = env.get::<Env>() {
                        env.output_id_map.insert(id, output.as_ref().id());
                        output.quick_assign(handle_output_event);
                    }
                },
                "wl_seat" if version >= 7 => {
                    let seat: Main<WlSeat> = registry.bind(version, id);
                    seat.quick_assign(handle_seat_event);
                },
                "zriver_status_manager_v1" => {
                    let status_manager = registry.bind(version, id);
                    if let Some(env) = env.get::<Env>() {
                        env.status_manager = Some(status_manager);
                    }
                },
                _ => {}
            }
        }
        GlobalEvent::Removed { id, interface: _ } => {
            if let Some(env) = env.get::<Env>() {
                if let Some(oid) = env.output_id_map.remove(&id) {
                    env.outputs.remove(&oid);
                }
            }
        }
    }
}

fn configuration() -> Flags {
    let mut default = Flags::default();
    let mut args = std::env::args();

    loop {
        match args.next() {
            Some(flag) => match flag.as_str() {
                "--seat" | "-s" => default.seat = args.next(),
                "--once" | "-o" => default.once = true,
                "--help" | "-h" => {
                    println!("Usage: ristate [option]\n");
                    println!("  --help | -h            print this message and exit");
                    println!("  --once | -o            print the current status and exit");
                    println!("  --seat | -s <string>   select the seat");
                    std::process::exit(0);
                }
                _ => {}
            },
            None => break,
        }
    }
    default
}

fn main() {
    let mut env = Env::new();

    let display = Display::connect_to_env().unwrap();
    let mut event_queue = display.create_event_queue();
    let attached_display = (*display).clone().attach(event_queue.token());

    GlobalManager::new_with_cb(
        &attached_display,
        global_manager_callback,
    );

    loop {
        event_queue
            .dispatch(&mut env, |event, object, _| {
                panic!(
                    "[callop] Encountered an orphan event: {}@{}: {}",
                    event.interface,
                    object.as_ref().id(),
                    event.name
                );
            })
            .unwrap();
        if !env.is_empty() && env.is_first_read_done() {
            env.fmt();
            if env.flags.once {
                std::process::exit(0);
            }
        }
    }
}
